
#include "aqmalloc.h"

/*
 Malloc works by manipulating 'chunks'. This is the name used by
 older malloc's, so it sticks. At any rate.
 A chunk has some information in it's header that is metadata for the
 chunk (size, etc.).
 
 These routines keep track of what pages are used so the garbage 
 collector will know what to search through. 
*/

/* 8/26/98 -

   See Bubl.h for the description of the bit fields.

   */

/*
  Type declarations
*/


struct malloc_chunk
{
    uint32        size;		/* Size in bytes of malloc (including this field) */
    struct malloc_chunk * fd;   /* double links -- used only if free.  */
    struct malloc_chunk * bk;
};

typedef struct malloc_chunk* mchunkptr;

typedef struct malloc_chunk * Chunk;


// This struct holds a begining page and the number of consecutive
// pages after it that are used after it.
struct pageRange
{
    char    * page;
    uint32    size;
};

struct pagesRecord
{
    struct pageRange next;	// This holds the pointer to the next page and the number of entries
    struct pageRange entry[681];
};


/*  sizes, alignments */

#define SIZE                   (4)
#define MALLOC_ALIGNMENT       (4*SIZE)
#define MALLOC_ALIGN_MASK      (MALLOC_ALIGNMENT - 1)
#define MINSIZE                (16)
/* probably 64k for a 4k page size */
#define MINPAGES               (16)

#define  PageSize 4096
#define  PageMask (PageSize - 1)



/* size field is or'ed with PREV_INUSE when previous adjacent chunk in use */

#define PREV_INUSE 0x1 

/* size field is or'ed with GC_INUSE when this chunk is garbage collected */

#define GC_INUSE 0x2

/* This indicates that the chunk is a forwarder for the real data -
   the ptr is in the 'fd' link. */

#define FORWARDER  (0x00000008)


/* Bits to mask off when extracting size */

#define SIZE_MASK (0x8000000F)
#define SIZE_BITS (~SIZE_MASK)


/* helpful macros */

#define ChunkSize(p)          ((p)->size & SIZE_BITS)

#define Chunk2Mem(p)          ((void *)((char *)(p) + SIZE))

#define Mem2Chunk(p)          ((void *)((char *)(p) - SIZE))







/* Globals */

static uint32               TopBytesLeft = 0;
static char               * Top = nil;

struct pagesRecord * Pages = nil;		// Root pointer to linked list of pages







/* 
  Debugging support 
*/

int (*mallocDebugHook) (int type,int mem) = NULL;  /* Debug Message Handler */




// Support routines


// This records what pages were used
// Each page holds entries consisting of 4 byte page pointer
// with 2 byte size.
// This stores the first page and then the number of pages
// in the setThe first two entries
// are the link to the next page and the count of entries used
// in that page
void recordPages(void * page, int numPages)
{
    struct pagesRecord * p;
    struct pagesRecord * pageBlock;
    struct pageRange   * pageEntry;
    uint32 * intPtr;
    uint32   i,j;
    
    

    // If this is the first time, get a page
    if (Pages == nil) {
	Pages = (struct pagesRecord *) calloc(PageSize, 1);
	//// We should throw a nasty exception here
	assert(Pages);
    }

    // Follow the chain to the last page
    p = Pages;

    while (p->next.page != nil) {
	p = (struct pagesRecord *) p->next.page;
    }
    

    // If there is room, record what pages can be recorded
    i = p->next.size;
    
    if (i == 682) {
	// Otherwise, allocate another page and add it to the list
	p->next.page = (char *)calloc(PageSize, 1);
	assert(p->next.page);
	i = 0;
	p = (struct pagesRecord *) p->next.page;
    }

    p->entry[i].page = page;
    p->entry[i].size = numPages;

    p->next.size++;
    
}



// This extends the top segment by 
uint32 addBytesTop(uint32 bytes)
{
    uint32   pagesNeeded;
    uint32   bytesNeeded;
    uint32   tmp;
    char   * p;
    char   * tmpTop;
    

    // Determine how many bytes are needed past page boundary
    bytesNeeded = bytes - TopBytesLeft;

    // Round up bytesNeeded to nearest page
    tmp = bytesNeeded + PageSize - 1;

    pagesNeeded = tmp / PageSize;

    if (pagesNeeded < MINPAGES) {
      pagesNeeded = MINPAGES;
    }
    
    // Ok grab pages needed
#ifdef __FreeBSD
    p = (char *) malloc(PageSize * pagesNeeded);
#else
    p = (char *) valloc(PageSize * pagesNeeded);
#endif

    if (!p) {
	return 0;
    }
    
    // round top up to the nearest page to see if it will match 'p'
    tmpTop = (char *) ((uint32)( Top + (PageSize - 1) ) & ~(PageSize - 1));

    // If they are equal, the pages were contigous to top and we can return
    // (This will happen in FreeBSD since it can allocate continous pages)
    if (p == tmpTop) {
	recordPages(p, pagesNeeded);
	TopBytesLeft += (PageSize * pagesNeeded);
	return 1;
    }
    
    // else, if pages needed still satisfies this request
    // we're done, just set the new topBytesLeft, the new top,
    // and record what pages were added
    if ((PageSize * pagesNeeded) >= bytes) {
	recordPages(p, pagesNeeded);
	TopBytesLeft = (PageSize * pagesNeeded);
	Top = p;
	return 1;
    }
    

    // If all of that fails, then recalc pages needed
    // without using the TopBytesLeft. This can be
    // accomplished through recursion
    
    // Otherwise free those pages, set the number of
    // bytes available in top to be 0, do a recursive
    // call.
    // and allocate more
    free(p);
    TopBytesLeft = 0;
    return (addBytesTop(bytes));
    
}




/* Main public routines */




/*
  Malloc Algorithm:

*/

void * aqmalloc(uint32 bytes)
{
  Chunk     chunk;


  /* Normalize the request size to a multiple of 16*/
  // So add 4 for the header(the size), then add 15 and divide by 16
  bytes = (bytes + 4 + MALLOC_ALIGN_MASK) & ~MALLOC_ALIGN_MASK;
  bytes = (bytes >= MINSIZE ? bytes : MINSIZE);

  /* Debug hook */
  if (mallocDebugHook) {
      mallocDebugHook(0, bytes);
  }


  // If we don't have space, allocate more
  if (TopBytesLeft < bytes) {
      if (!addBytesTop(bytes)) {
	  return 0;
      }
  } 

  // Now give the space to the chunk
  chunk = (Chunk) Top;
  Top += bytes;
  chunk->size = bytes;
  //// Need to set the in_use bit
  TopBytesLeft -= bytes;
  
  return Chunk2Mem(chunk);
}


/*

  free() algorithm :

*/


void aqfree(void * memory)
{
    /*
    Chunk chunk;

    // aqfree(0) has no effect
    if (chunk == 0) {
      return;
    }

    chunk = Mem2chunk(memory);

    */
}


/*

  calloc calls malloc, then zeroes out the allocated chunk.

*/

void * aqcalloc(uint32 bytes)
{
    Chunk chunk;
    uint32 size;

    
    void * memory = aqmalloc(bytes);

    if (memory == 0) {
	return 0;
    } else {
	chunk = Mem2Chunk(memory);
    }
    
    size = ChunkSize(chunk);

    MALLOC_ZERO(memory, size - SIZE);
    return memory;
}

void *  aqrealloc(void * mem, uint32 bytes)
{
    Chunk chunk;
    uint32 size;

    // Get some new memory
    void * memory = aqmalloc(bytes);

    if (memory == 0) {
	return 0;
    }

    if (mem == (void *) 0) {
	return memory;
    }
    

    // Get the size of the old memory
    chunk = Mem2Chunk(mem);
    size = ChunkSize(chunk);

    // Copy the old memory into the new memory
    MALLOC_COPY(memory, mem, size);
    
    return memory;
}
